#!/usr/bin/env python3
# -*- coding:utf-8 -*-

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

"""
(c) 2020 Miguel Colom
http://mcolom.info

Under GNU GPL license. See LICENCE file for details.
"""

import sys
import os
import urllib.request
import datetime
import time
import argparse
import json
from os.path import join, isfile
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
import numpy as np
from numpy import array, sqrt, median, mean, std, insert, add, arange, ones, roll
import pandas as pd
from scipy.stats import linregress
from dataclasses import dataclass

@dataclass
class Constants:
    a: float = 1
    b: float = 5
    tau: int = 7
    w: int = 7
    lambs: object = None

class Functions:
    get_Rn_curve: object = None
    get_Rn_r_curve: object = None
    get_conf_interval: object = None

def cori_numerator(t, v):
    """
    Compute Cori's method numerator
    """    
    # First index s = 0: t-tau+1 = 0 ==> t = tau - 1
    return Constants.a + sum([(v[s]) for s in range(t-Constants.tau+1, t+1)])

def cori_denominator(t, v):
    """
    Compute Cori's method denominator
    """    
    # First index s-k = 0: s-k=0; t-tau+1-k=0; t = tau + len(Constants.lambs) - 1
    return 1/Constants.b +  sum([v[s-k]*Constants.lambs[k] for s in range(t-Constants.tau+1, t+1) for k in range(len(Constants.lambs))])

def save_country(cases, deaths, start_date, end_date, population, country, data_directory):
    '''
    Save the cases of one particular country
    '''
    filename = join(data_directory, f"cases_{country}.txt")
    with open(filename, "wt") as f:
        for num_cases in cases:
            f.write(f"{num_cases}\n")

    filename = join(data_directory, f"deaths_{country}.txt")
    with open(filename, "wt") as f:
        for num_deaths in deaths:
            f.write(f"{num_deaths}\n")

    filename = join(data_directory, f"start_date_{country}.txt")
    with open(filename, "wt") as f:
        f.write(f"{start_date.timestamp()}\n")

    filename = join(data_directory, f"population_{country}.txt")
    with open(filename, "wt") as f:
        f.write(f"{population}\n")

def date_range_num(start_date, num_dates):
    '''
    Return a range of dates, from start_date with cardinal num_dates
    '''
    dates = []
    date = start_date

    for _ in range(num_dates):
        dates.append(date)
        date += datetime.timedelta(days=1)
    return dates

def date_range(start_date, end_date):
    '''
    Return a range of dates, from start_date to end_date, both included
    '''
    dates = []
    date = start_date
    while date <= end_date:
        dates.append(date)
        date += datetime.timedelta(days=1)
    return dates


def demo_failure(message):
    '''
    Exit with a demo_failure message
    '''
    print(message)
    with open("demo_failure.txt", "wt") as f:
        f.write(message)
        f.write("\n")
    sys.exit(0)

def read_daily(filename):
    '''
    Read the file with the cases
    '''
    try:
        with open(filename, "rt") as fin:
            lines = fin.readlines()

        cases = []
        for line in lines:
            if line:
                cases.append(float(line))
    except (ValueError, UnicodeDecodeError):
        demo_failure("Uploaded data: bad format.")

    return array(cases)

def read_float(filename):
    '''
    Read the file with the cases
    '''
    with open(filename, "rt") as fin:
        return float(fin.readline())

def read_country(country, data_directory):
    '''
    Read a country: deaths, infections, starting date, ...
    '''
    # Deaths
    filename = join(data_directory, f"deaths_{country}.txt")
    d = read_daily(filename)

    # Infections
    filename = join(data_directory, f"cases_{country}.txt")
    i = read_daily(filename)

    # Population
    filename = join(data_directory, f"population_{country}.txt")
    P = int(read_float(filename))

    # Dates
    filename = join(data_directory, f"start_date_{country}.txt")
    start_timestamp = read_float(filename)

    start_date = datetime.datetime.fromtimestamp(start_timestamp)
    dates = date_range_num(start_date, len(i))

    assert len(dates) == len(d) == len(i)
    return d, i, P, dates

def country_code_to_name(code):
    '''
    Convert a country code into a name
    Ex: United_States_of_America --> United States of America
    '''
    return code.replace("_", " ")

def get_Rn_r_Cori(t, d):
    '''
    Compute R_n with Cori's method
    '''
    l = len(Constants.lambs)
    #assert n >= l
    
    numerator = cori_numerator(t, d)
    denominator = cori_denominator(t, d)

    if denominator < 1e-9: # Avoid numerical problems
        return 0
    
    # Avoid values close to zero
    R = numerator / (denominator +  5/(1+denominator))
    R = min(R, 3.5)
    return R

def get_Rn_r_ERN(n, d):
    '''
    Compute R_n with Cori's method
    '''
    assert n >= len(Constants.lambs)
    assert n < len(d)

    # Compute sum
    l = len(Constants.lambs)
    denominator = sum([Constants.lambs[k] * d[n-k-1] for k in range(l)])

    if denominator < 1e-9: # Avoid numerical problems
        return 0

    # Avoid values close to zero
    R = d[n] / (denominator +  5/(1+denominator))
    R = min(R, 3.5)
    return R

def get_Rn_Cori(t, i):
    '''
    Compute R_n with Cori's method
    '''
    l = len(Constants.lambs)
    #assert n >= l
    
    numerator = cori_numerator(t, i)
    denominator = cori_denominator(t, i)

    if denominator < 1e-9: # Avoid numerical problems
        return 0
    
    # Avoid values close to zero
    R = numerator / (denominator +  5/(1+denominator))
    R = min(R, 3.5)
    return R

def get_Rn_ERN(n, i):
    '''
    Compute R_n
    '''
    l = len(Constants.lambs)
    assert n >= l

    denominator = sum([Constants.lambs[k] * i[n-k-1] for k in range(l)])

    if denominator < 1e-9: # Avoid numerical problems
        return 0

    # Avoid values close to zero
    R = i[n] / (denominator +  5/(1+denominator))
    R = min(R, 3.5)
    return R

def filtering(C, operator, w):
    '''
    Median filter. The last values are lost
    '''
    l1 = int((w - 1) / 2)
    l2 = w - l1

    assert len(C) >= w
    C_filtered = [operator(C[k-l1:k+l2]) for k in range(l1, len(C)-l1)]
    C_filtered = array(C_filtered)

    # Add heading values copied from C
    for i in range(l1):
        C_filtered = insert(C_filtered, 0, C[l1 - i - 1])
    return C_filtered


def do_save_plot(filename, Rn_minus_r, Rn, err_Rn_minus_r, err_Rn, s, start_date, end_date, country):
    '''
    Do save the plot
    '''
    fig, ax = plt.subplots()


    if start_date and end_date:
        # Dates available: date plot
        num_dates = len(Rn_minus_r)

        fig.autofmt_xdate()
        dates = date_range_num(start_date, num_dates)

        if Rn_minus_r is not None:
            shifted_dates = add(dates[0:len(Rn_minus_r)], datetime.timedelta(days=-s))
            ax.plot_date(shifted_dates, Rn_minus_r, ls='-', marker='.', label="$R_{n-r}$")
            ax.fill_between(shifted_dates, Rn_minus_r-err_Rn_minus_r, Rn_minus_r+err_Rn_minus_r, alpha=0.3)

        if Rn is not None:
            ax.plot_date(dates[0:len(Rn)], Rn, ls='-', marker='.', label="$R_n$")
            ax.fill_between(dates[0:len(Rn)], Rn-err_Rn, Rn+err_Rn, alpha=0.3)
    else:
        # No dates given: normal numerical plot
        dates = start_date + arange(num_dates)
        ax.plot_date(shifted_dates, Rn_minus_r, ls='-', marker='.', label="$R_{n-r}$")

    ax.axhline(y=1, ls="--", color="black")
    ax.set_ylim((0, 3.5))

    country_name = country_code_to_name(country)
    ax.set_title(country_name)
    ax.set_ylabel("ERN")
    ax.legend()
    ax.grid()
    plt.savefig(filename)

def save_deaths_infections_plot(filename, d, i, d_f, i_f, dates, country):
    '''
    Deaths and infections plot
    '''
    plt.close("all")
    fig, ax1 = plt.subplots()

    num_dates = len(d)
    assert num_dates == len(i)

    # Skip the first points in the plot
    q = min(90, num_dates)
    skip = len(d) - q
    #
    d = d[skip:]
    i = i[skip:]
    d_f = d_f[skip:]
    i_f = i_f[skip:]
    dates = dates[skip:]

    fig.autofmt_xdate()

    alpha = 0.4
    orange = '#ff7f0e'
    blue = '#1f77b4'

    ax1.plot_date(dates, d, ls='-', marker='.', label="$d$", color=blue, alpha=alpha)
    ax1.plot_date(dates[0:len(d_f)], d_f, ls='-', marker='', linewidth=2, label="$d_f$", color=blue)
    ax1.set_ylim(bottom=0)
    ax1.tick_params(axis='y', labelcolor=blue)

    ax2 = ax1.twinx()  # instantiate a second axes that shares the same x-axis

    ax2.plot_date(dates, i, ls='-', marker='.', label="$i$", color=orange, alpha=alpha)
    ax2.plot_date(dates[0:len(i_f)], i_f, ls='-', marker='', linewidth=2, label="$i_f$", color=orange)
    ax2.set_ylim(bottom=0)
    ax2.tick_params(axis='y', labelcolor=orange)

    country_name = country_code_to_name(country)
    ax1.set_title(f"Infections and deaths in {country_name}\nOrange: infections. Blue: deaths.")
    ax1.set_ylabel("Counts")
    #ax1.legend()
    #ax2.legend()
    ax1.grid()
    plt.savefig(filename)

def get_Rn_r_curve_Cori(d):
    '''
    Compute the R_{n-r} curve for all valid n
    '''
    start, stop = Constants.tau + len(Constants.lambs) - 1, len(d)
    Rn_minus_r = [get_Rn_r_Cori(t, d) for t in range(start, stop)]
    return array(Rn_minus_r)

def get_Rn_r_curve_ERN(d):
    '''
    Compute the R_{n-r} curve for all valid n
    '''
    start, stop = len(Constants.lambs), len(d)
    Rn_minus_r = [get_Rn_r_ERN(n, d) for n in range(start, stop)]
    return array(Rn_minus_r)


def get_Rn_curve_Cori(i):
    '''
    Compute the R_n curve for all valid n
    '''
    start, stop = Constants.tau + len(Constants.lambs) - 1, len(i)
    Rn = [get_Rn_Cori(t, i) for t in range(start, stop)]
    return array(Rn)

def get_Rn_curve_ERN(i):
    '''
    Compute the R_n curve for all valid n
    '''
    start, stop = len(Constants.lambs), len(i)
    Rn = [get_Rn_ERN(n, i) for n in range(start, stop)]
    return array(Rn)

    
def save_plot(Rn_minus_r, Rn, dates, s, country, filename="plot.png"):
    '''
    Save the Rn_minus_r and Rn plot of the last days
    '''
    if dates is None:
        start_date, end_date = None, None
    else:
        dates = list(dates)
        start_date = min(dates)
        end_date = max(dates)

    err_Rn_minus_r = Functions.get_conf_interval(Rn_minus_r)
    err_Rn = Functions.get_conf_interval(Rn)

    # Show only the last q days
    q = min(90, len(Rn))
    Rn_minus_r = Rn_minus_r[-q:]
    Rn = Rn[-q:]

    err_Rn_minus_r = err_Rn_minus_r[-q:]
    err_Rn = err_Rn[-q:]

    if start_date:
        start_date = end_date - datetime.timedelta(days=q-1)

    do_save_plot(filename, Rn_minus_r, Rn, err_Rn_minus_r, err_Rn, s, start_date, end_date, country)

def get_conf_interval_ERN(i_f):
    '''
    RMSE between the last w points regression and the data.
    '''
    w = Constants.w

    errors = []
    for k in range(w-1):
        errors.append(0)

    for k in range(w-1, len(i_f)):
        data = i_f[k-w+1 : k+1]
        res = linregress(range(w), data)

        regressed = [res.slope * q + res.intercept for q in range(w)]

        RMSE = 2 * std(data - regressed)
        errors.append(RMSE)

    return array(errors)

def get_conf_interval_Cori(i):
    '''
    Obtain a error for the confidence interval of Cori's method
    '''
    start, stop = Constants.tau + len(Constants.lambs) - 1, len(i)

    errors = []
    for k in range(Constants.tau + len(Constants.lambs) - 1):
        errors.append(0)

    for t in range(start, stop):
        numerator = cori_numerator(t, i)
        denominator = cori_denominator(t, i)

        errors.append(sqrt(numerator) / denominator)
    return array(errors)

def aggregate_countries(agg_countries):
    '''
    Obtain aggreagated deaths, infections, and population of a group of
    countries
    '''
    agg_dates = []
    agg_population = 0

    # Read every country
    d, i, P, dates = {}, {}, {}, {}
    for c in agg_countries:
        c_d, c_i, c_P, c_dates = read_country(c, data_directory)
        d[c] = c_d
        i[c] = c_i
        P[c] = c_P
        dates[c] = c_dates
        
    # Look for start and end dates for all countries
    agg_start_date = max([min(dates[c]) for c in agg_countries])
    agg_end_date = min([max(dates[c]) for c in agg_countries])

    agg_dates = date_range(agg_start_date, agg_end_date)
    agg_deaths = np.zeros(len(agg_dates))
    agg_cases = np.zeros(len(agg_dates))
    
    for idx_agg, date in enumerate(agg_dates):
        for c in agg_countries:            
            if date in dates[c]:
                idx_c = dates[c].index(date)
                agg_deaths[idx_agg] += d[c][idx_c]
                agg_cases[idx_agg] += i[c][idx_c]

    for c in agg_countries:
        agg_population += P[c]

    return agg_deaths, agg_cases, agg_population, agg_dates


def update_in_background(data_directory, background=True):
    """
    Update countrie's data in background
    """
    program_path = "./background_updater.py" if isfile("./background_updater.py") else "background_updater.py"
    command = f'{program_path} --data_directory {data_directory}'
    if background:
        command += " &"
    print(f"Running: {command}")
    os.system(command)


# Parse arguments
parser = argparse.ArgumentParser()
parser.add_argument("country", help="Country name")
parser.add_argument("--s", type=int, default=6, help="Delay")
parser.add_argument("--lambdas_calculation", type=str, default="nishiura", help="Lambdas values")
parser.add_argument("--method", type=str, default="ERN", help="Method to compute R")
parser.add_argument("--filter_name", type=str, default="median", help="How to filter the cases/deaths input")
parser.add_argument("--data_directory", help="Data directory", default="daily_data/")
args = parser.parse_args()

# Copy arguments
data_directory = args.data_directory
s = args.s
lambdas_calculation = args.lambdas_calculation
method = args.method
filter_name = args.filter_name
country = args.country

# Set functions depending on the method
if method == 'ERN':
    Functions.get_Rn_r_curve = get_Rn_r_curve_ERN
    Functions.get_Rn_curve = get_Rn_curve_ERN
    Functions.get_conf_interval = get_conf_interval_ERN
elif method == 'Cori':
    Functions.get_Rn_r_curve = get_Rn_r_curve_Cori
    Functions.get_Rn_curve = get_Rn_curve_Cori
    Functions.get_conf_interval = get_conf_interval_Cori
else:
    raise ValueError(method)

# Update countries in foreground if local storage is empty
if not isfile(join(data_directory, 'cases_FRA.txt')):
    update_in_background(data_directory, background=False)

if lambdas_calculation == 'world':
    # Coefficients of coronavirus transmission SARS-2019, World, calibrated.
    # See Tahar's article.
    Constants.lambs = array((0.1176, 0.1103, 0.1029, 0.0956, 0.0882, 0.0809, 0.0735, 0.0662, 0.0588, 0.0515, 0.0441, 0.0368, 0.0294, 0.0221, 0.0147, 0.0074))
elif lambdas_calculation == 'nishiura':
    # Serial interval
    Constants.lambs = array((1.5, 4, 6.25, 5.5, 2.25, 1, 1.5, 1.75, 2.25, 1.5))
#elif lambdas_calculation == 'incubation':
    # P. Abry et al. (main_v3.pdf), Figure 1, page 8
    # ref 16, page 20 --> (A) Incubation period (days)
    # Constants.lambs = array((32, 440, 522, 728, 1290, 794, 794, 918, 794, 778, 644, 454, 526, 284, 268, 218, 112, 168, 30, 0, 48, 14, 32, 10))
elif lambdas_calculation in 'ma':
    # P. Abry et al. (main_v3.pdf), Figure 1, page 8
    # ref 16, page 20 --> (C) Serial interval (days)
    # Constants.lambs = array((30, 70, 130, 40, 114,       464, 550, 724, 927, 900, 637, 840, 725, 580, 565, 477, 450, 260, 246, 260, 230, 245, 215, 100, 55, 45, 45, 15, 15, 15))
    # We accumulate negative times days in [0] and remove them
    Constants.lambs = array((848, 550, 724, 927, 900, 637, 840, 725, 580, 565, 477, 450, 260, 246, 260, 230, 245, 215, 100, 55, 45, 45, 15, 15, 15))
elif lambdas_calculation == 'du':
    # Article The serial interval of COVID-19 from publicly reported confirmed cases. Running Head: The serial interval of COVID-19.
    # Authors: Zhanwei Du et al.
    # We accumulate negative times days in [0] and remove them
    # Constants.lambs = array((15, 42, 15, 15, 0, 40, 102, 168, 150, 150,    540, 1026, 873, 1068, 783, 1068, 636, 702, 486, 429, 342, 342, 302, 84, 130, 130, 150, 81, 63, 0, 24))
    Constants.lambs = array((1237, 1026, 873, 1068, 783, 1068, 636, 702, 486, 429, 342, 342, 302, 84, 130, 130, 150, 81, 63, 0, 24))
else:
    demo_failure(f"Unknown calculation technique for lambdas: '{lambdas_calculation}'")

# lambs normalization
Constants.lambs = Constants.lambs / sum(Constants.lambs)


is_aggregated = country.startswith("agg_")
if is_aggregated:
    if country == "agg_EU":
        #agg_countries = ("Austria", "Belgium", "Bulgaria", "Croatia", "Cyprus", "Czechia", "Germany", "Denmark", "Estonia", "Finland", "France", "Greece", "Hungary", "Ireland", "Italy", "Latvia", "Lithuania", "Luxembourg", "Malta", "Netherlands", "Poland", "Portugal", "Romania", "Serbia", "Slovakia", "Slovenia", "Sweden", "Spain")
        agg_countries = ("AUT", "BEL", "BGR", "HRV", "CYP", "CZE", "DEU", "DNK", "EST", "FIN", "FRA", "GRC", "HUN", "IRL", "ITA", "LVA", "LTU", "LUX", "MLT", "NLD", "POL", "PRT", "ROU", "SRB", "SRB", "SVN", "SWE", "ESP")
    elif country == "agg_south_east_asia":
        #agg_countries = ("Singapore", "Malaysia", "Indonesia", "Brunei_Darussalam", "Thailand", "Vietnam")
        agg_countries = ("SGP", "MYS", "IDN", "BRN", "THA", "VNM")
    else:
        demo_failure(f"Unknown aggregated '{country}'")

# Extract from reports daily data
# d: deaths
# i: infections
# dates: dates
# P: population
d, i, P, dates = aggregate_countries(agg_countries) if is_aggregated else read_country(country, data_directory)

# Filter d and i with the median. The last 3 points will be lost
if filter_name == 'median':
    filter_operator = median
elif filter_name == 'mean':
    filter_operator = mean
else:
    raise ValueError(filter_name)
#
if method == 'Cori':
    filter_operator = mean

d_f = filtering(d, filter_operator, Constants.w)
i_f = filtering(i, filter_operator, Constants.w)
save_deaths_infections_plot("deaths_and_infections.png", d, i, d_f, i_f, dates, country)

if method == 'ERN':
    filter_days_removed = int((Constants.w-1)/2)    
elif method == 'Cori':
    d_f = d
    i_f = i
    assert Constants.tau % 2 == 1
    #filter_days_removed = int((Constants.tau-1)/2)
    filter_days_removed = 0
else:
    raise ValueError(method)

if filter_days_removed > 0:
    dates = dates[:-filter_days_removed] # Remove days because of the median filter

# Compute Rn_minus_r and Rn curves
Rn_minus_r = Functions.get_Rn_r_curve(d_f)
Rn = Functions.get_Rn_curve(i_f)

# And save plot
dates = dates[len(dates) - len(Rn):]
save_plot(Rn_minus_r, Rn, dates, s, country)

# Update countries in background
update_in_background(data_directory)
